/* C header created by blob2c from input file fenders.bin */

#ifndef fenders_bin_H
#define fenders_bin_H

extern unsigned char fenders_bin[];
extern int fenders_bin_len;

#endif  /* fenders_bin_H */
