/* BXL and BXE token tables are identical up to 0x56 */
const char *bxe_cmds[] = {
	"LOCAL",     /* $57 */
	"EXTEND",    /* $58 */
	"PROCEDURE", /* $59 */
	0,           /* $5A */
	"CALL",      /* $5B */
	"SORTUP",    /* $5C */
	"SORTDOWN",  /* $5D */
	"EXIT",      /* $5E */
	"NUM",       /* $5F */
	"HITCLR",    /* $60 */
	"INVERSE",   /* $61 */
	"NORMAL",    /* $62 */
	"BLOAD",     /* $63 */
	"BSAVE"      /* $64 */
};

const int bxe_cmd_size = sizeof(bxe_cmds);
