/* this stuff is in a separate .h file so that eventually,
   listbas.c can include it (it'll execute "whichbas -s" to detect
   the BASIC type). */

#define SRET_ATARI                 3
#define SRET_TURBO                 4
#define SRET_BXL                   5
#define SRET_BXE                   6
#define SRET_TURBO_OR_BXL          7
#define SRET_TURBO_OR_BXE          8
#define SRET_TURBO_OR_BXL_OR_BXE   9
#define SRET_UKNOWN_DERIV         10
#define SRET_AMSB                 11
#define SRET_EXTENDED_BXE         12
#define SRET_COMPILED_TURBO       13
#define SRET_APLUS                14
#define SRET_NOT_BASIC            64
