/* Copyright 1997 Ken Siders */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "atr.h"
#include "atdos.h"


/********************************************************************
 Testing 
********************************************************************/


int main( int argc, char **argv)
   {
   long size = 0;

   printf("MakeAtr Version 0.9u (c)1997 Ken Siders\n");
	printf("Ported and modified by B. Watson, 2007\n");
   printf("This program may be freely distributed\n\n");


   if (argc != 3)
      {
      printf("usage: unmakatr atrname.atr file\n\n");
      }
   else
      {
      size = ExtractExeFromBootAtr( argv[1], argv[2] );
      if ( size <= 0 ) {
         printf("Error #%ld encountered in extracting\n\n", -size);
			exit(size);
		} else {
         printf("No errors, %s extracted successfully (%ld bytes)\n\n", argv[2], size);
		}
	}

	exit(0);

}


