/* C header created by blob2c from input file loadscreen.bin */

#ifndef loadscreen_bin_H
#define loadscreen_bin_H

extern unsigned char loadscreen_bin[];
extern int loadscreen_bin_len;

#endif  /* loadscreen_bin_H */
