#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "bas.h"
#include "bcdfp.h"

typedef struct {
	unsigned short lineno;
	unsigned short pos;
	unsigned char cmd;
} lineref_t;

extern lineref_t *linerefs[];
extern int refcounts[];
extern int lines_exist[];

extern void add_lineref(unsigned short from, unsigned short pos);
extern void build_ref_table(void);
extern void free_ref_table(void);
