/* these start at token 0x38 */
const char *turbo_cmds[] = {
	"DPOKE", /* $38 */
	"MOVE",
	"-MOVE",
	"*F",
	"REPEAT",
	"UNTIL",
	"WHILE",
	"WEND",
	"ELSE",  /* $40 */
	"ENDIF",
	"BPUT",
	"BGET",
	"FILLTO",
	"DO",
	"LOOP",
	"EXIT",
	"DIR",   /* $48 */
	"LOCK",
	"UNLOCK",
	"RENAME",
	"DELETE",
	"PAUSE",
	"TIME$=",
	"PROC",
	"EXEC",   /* $50 */
	"ENDPROC",
	"FCOLOR",
	"*L",
	"------------------------------",
	"RENUM",
	"DEL",
	"DUMP",
	"TRACE",  /* $58 */
	"TEXT",
	"BLOAD",
	"BRUN",
	"GO#",
	"#",
	"*B",
	"PAINT",
	"CLS",    /* $58 */
	"DSOUND",
	"CIRCLE",
	"%PUT",
	"%GET"    /* $64 */
};

const int turbo_cmd_size = sizeof(turbo_cmds);

/* these start at token 0x55 */
const char *turbo_ops[] = {
	"DPEEK", /* $55 */
	"&",
	"!",
	"INSTR", /* $58 */
	"INKEY$",
	" EXOR ",
	"HEX$",
	"DEC",
	" DIV ",
	"FRAC",
	"TIME$",
	"TIME",  /* $60 */
	" MOD ",
	" EXEC ",
	"RND",
	"RAND",
	"TRUNC",
	"%0",
	"%1",
	"%2",    /* $68 */
	"%3",
	" GO# ",
	"UINSTR",
	"ERR",
	"ERL",   /* $6D */
};

const int turbo_ops_size = sizeof(turbo_ops);

