/* C header created by blob2c from input file fendersdbl.bin */

#ifndef fendersdbl_bin_H
#define fendersdbl_bin_H

extern unsigned char fendersdbl_bin[];
extern int fendersdbl_bin_len;

#endif  /* fendersdbl_bin_H */
