//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#define PRG_NAME	"aext"	
#define PRG_VERSION	"1.02"
#define PRG_COPYRIGHT	"1998-2002"
#define PRG_UNDERGPL	"1"
#define PRG_AUTHOR	"Jindrich Kubec <kubecj@asw.cz>"

#define PRG_URL 	"http://www.asw.cz/~kubecj"

#define PRG_DESC	"Atari file extractor."
#define PRG_USAGE	"[options] infile [outfile]"

#define PRG_ARCH	"PC"
#define PRG_TOOLS	"PERL"

#define PRG_SRCS	"!GENERATED!"
#define PRG_BINS	"rel/aext.exe"
