atr, a
output is ATR
=g_dtypeOut = DISK_ATR;
=

xfd, x
output is XFD
=g_dtypeOut = DISK_XFD;
=

scp, s
output is SCP
=g_dtypeOut = DISK_SCP;
=g_bClassic = TRUE;
=

dcm, d
output is DCM
=g_dtypeOut = DISK_DCM;
=g_bClassic = TRUE;
=

long, classic
output has 'classic' size
=g_bClassic = TRUE;
=

message, mess, text
message to display while loading
text=SWFN_GETPATH( g_szMessage );
=

