//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#ifndef __ERRBASE_H__
#define __ERRBASE_H__

#define CDISK_BASE	0x1000
#define CATR_BASE		0x2000
#define CDCM_BASE		0x3000
#define CSCP_BASE		0x4000
#define CXFD_BASE		0x5000
#define CDI_BASE		0x6000

#endif
